with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics,
  ada.numerics.long_elementary_functions,
  terms;

procedure cylinder2 is

  type vec3 is array (1 .. 3) of long_float;

  type elem1_rec is
    record
      glob : terms.imat1 (1 .. 2);
    end record;

  type elem1_vec is array (integer range <>) of elem1_rec;

  type elem1_ptr is access elem1_vec;

  type elem2_rec is
    record
      glob : terms.imat1 (1 .. 3);
    end record;

  type elem2_vec is array (integer range <>) of elem2_rec;

  type elem2_ptr is access elem2_vec;

  ni,nj,nk,nn,state : integer;
  radius,height     : long_float;
  area,young,value  : long_float;
  stiffness,poisson : long_float;
  xo,yo,zo          : terms.fmat1_ptr;
  elem1             : elem1_ptr;
  elem2             : elem2_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  function node (i,j : in integer) return integer is

  begin
    return ni * j + (i mod ni) + 1;
  end node;

  procedure vertex (i,j : in integer;
                    x   : out vec3) is

    use
      ada.numerics,
      ada.numerics.long_elementary_functions;

    t : long_float;

  begin
    t := 2.0 * pi / long_float(ni) * long_float(i mod ni);
    x(1) := radius * cos(t);
    x(2) := radius * sin(t);
    x(3) := (long_float(j) / long_float(nj) - 0.5) * height;
  end vertex;

  procedure initia is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    i,j,glob,e,ne : integer;
    x             : vec3;
    io            : file_type;

  begin
    open(io, in_file, "cylinder2.txt");

    -- The circumference is divided in ni parts
    -- The height is divided in nj parts
    -- The numbers ni and nj must be even
    -- The diagonals are placed skipping nk nodes between top and bottom circumferences
    -- The number of diagonals is equal to ni/(nk + 1)

    scan(io);
    get(io,radius);
    get(io,ni);
    get(io,height);
    get(io,nj);
    skip_line(io);

    scan(io);
    get(io,nk);
    get(io,area);
    get(io,young);
    get(io,state);
    get(io,value);
    skip_line(io);

    scan(io);
    get(io,stiffness);
    get(io,poisson);
    skip_line(io);

    close(io);

    -- nn
    nn := ni * (nj + 1);

    -- xo, yo, zo
    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    -- node coordinates
    for j in 0 .. nj loop
      for i in 0 .. ni - 1 loop
        vertex(i,j,x);
        glob := node(i,j);
        xo(glob) := x(1);
        yo(glob) := x(2);
        zo(glob) := x(3);
      end loop;
    end loop;

    -- elements 1
    ne := ni / (nk + 1);
    elem1 := new elem1_vec (1 .. ne);

    for e in elem1'range loop
      i := (e - 1) * (nk + 1);
      elem1(e).glob(1) := node(i,0);
      elem1(e).glob(2) := node(i + nk + 1,nj);
    end loop;

    -- elements 2
    ne := 2 * ni * nj;
    elem2 := new elem2_vec (1 .. ne);

    e := 0;

    j := 1;
    while j < nj loop

      i := 1;
      while i < ni loop
        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i + 1,j);
        elem2(e).glob(3) := node(i + 1,j + 1);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i + 1,j + 1);
        elem2(e).glob(3) := node(i,j + 1);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i,j + 1);
        elem2(e).glob(3) := node(i - 1,j + 1);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i - 1,j + 1);
        elem2(e).glob(3) := node(i - 1,j);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i - 1,j);
        elem2(e).glob(3) := node(i - 1,j - 1);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i - 1,j - 1);
        elem2(e).glob(3) := node(i,j - 1);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i,j - 1);
        elem2(e).glob(3) := node(i + 1,j - 1);

        e := e + 1;
        elem2(e).glob(1) := node(i,j);
        elem2(e).glob(2) := node(i + 1,j - 1);
        elem2(e).glob(3) := node(i + 1,j);

        i := i + 2;
      end loop;

      j := j + 2;
    end loop;
  end initia;

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    maxls,maxminor,maxlist : integer;
    lstol,mntol,mnref      : long_float;
    io                     : file_type;

  begin
    -- solution parameters
    lstol := 1.0E-01;
    maxls := 20;
    mntol := 1.0E-06;
    mnref := 1.0;
    maxminor := 30 * nn;
    maxlist := 3;

    create(io, out_file, "input.txt");

    put(io,"        lstol     maxls             mntol             mnref  maxminor   maxlist");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,lstol,2,6,3);
    put(io,"     ");
    put(io,maxls,5);
    put(io,"     ");
    put(io,mntol,2,6,3);
    put(io,"     ");
    put(io,mnref,2,6,3);
    put(io,"     ");
    put(io,maxminor,5);
    put(io,"     ");
    put(io,maxlist,5);
    new_line(io);

    -- node coordinates
    new_line(io);
    new_line(io);
    put(io," node           coord x           coord y           coord z");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,nn,5);
    new_line(io);

    new_line(io);
    for glob in 1 .. nn loop
      put(io,glob,5);
      put(io,"     ");
      put(io,xo(glob),2,6,3);
      put(io,"     ");
      put(io,yo(glob),2,6,3);
      put(io,"     ");
      put(io,zo(glob),2,6,3);
      new_line(io);
    end loop;

    -- elements 1
    new_line(io);
    new_line(io);
    put(io," elem      node      node              area             young");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,elem1'last,5);
    new_line(io);

    new_line(io);
    for e in elem1'range loop
      put(io,e,5);
      for loc in 1 .. 2 loop
        put(io,"     ");
        put(io,elem1(e).glob(loc),5);
      end loop;
      put(io,"     ");
      put(io,area,2,6,3);
      put(io,"     ");
      put(io,young,2,6,3);
      new_line(io);
    end loop;

    -- elements 2
    new_line(io);
    new_line(io);
    put(io," elem      node      node      node         stiffness           poisson");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,elem2'last,5);
    new_line(io);

    new_line(io);
    for e in elem2'range loop
      put(io,e,5);
      for loc in 1 .. 3 loop
        put(io,"     ");
        put(io,elem2(e).glob(loc),5);
      end loop;
      put(io,"     ");
      put(io,stiffness,2,6,3);
      put(io,"     ");
      put(io,poisson,2,6,3);
      new_line(io);
    end loop;

    -- known displacements
    new_line(io);
    new_line(io);
    put(io," node      axis             displ");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,"    0");
    new_line(io);

    -- set cable element state
    new_line(io);
    new_line(io);
    case state is
      when 1 =>
        put(io," elem     state               cut");
        new_line(io);

      when 2 =>
        put(io," elem     state           tension");
        new_line(io);

      when others =>
        null;
    end case;
    put(io,":");
    new_line(io);
    put(io,elem1'last,5);
    new_line(io);

    new_line(io);
    for e in elem1'range loop
      put(io,e,5);
      put(io,"     ");
      put(io,state,5);
      put(io,"     ");
      put(io,value,2,6,3);
      new_line(io);
    end loop;

    -- loading
    new_line(io);
    new_line(io);
    put(io," node      axis             force");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,"    0");
    new_line(io);

    close(io);
  end input;

  procedure autocad is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    glob : integer;
    io   : file_type;

  begin
    create(io, out_file, "cylinder2.scr");

    -- set running object snap mode to NONE
    put(io,"osmode 0");
    new_line(io);

    -- elements 2
    put(io,"layer make Elem2 ");
    new_line(io);

    for e in elem2'range loop
      put(io,"3dface ");
      for loc in 1 .. 3 loop
        glob := elem2(e).glob(loc);
        put(io,xo(glob),1,8,3);
        put(io,",");
        put(io,yo(glob),1,8,3);
        put(io,",");
        put(io,zo(glob),1,8,3);
        put(io," ");
      end loop;
      put(io," ");
      new_line(io);
    end loop;

    -- elements 1
    put(io,"layer make Elem1 ");
    new_line(io);

    put(io,"-color");
    new_line(io);
    put(io,"green");
    new_line(io);

    for e in elem1'range loop
      put(io,"3dpoly ");
      for loc in 1 .. 2 loop
        glob := elem1(e).glob(loc);
        put(io,xo(glob),1,8,3);
        put(io,",");
        put(io,yo(glob),1,8,3);
        put(io,",");
        put(io,zo(glob),1,8,3);
        put(io," ");
      end loop;
      new_line(io);
    end loop;

    -- autocad
    put(io,"-color");
    new_line(io);
    put(io,"white");
    new_line(io);

    put(io,"layer set 0 ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad;

begin
  initia;
  input;
  autocad;
end cylinder2;
