with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics.long_elementary_functions,
  terms,
  nlp;

procedure tensegrity is

  subtype fmat3x1 is terms.fmat1 (1 .. 3);

  subtype fmat9x1 is terms.fmat1 (1 .. 9);

  subtype fmat2x3 is terms.fmat2 (1 .. 2, 1 .. 3);

  type cable_rec is
    record
      area    : long_float;
      young   : long_float;
      state   : integer := 0;
      cut     : long_float := 0.0;
      tension : long_float := 0.0;
      glob    : terms.imat1 (1 .. 2);
      length  : long_float;
      u       : fmat3x1;
    end record;

  type cable_vec is array (integer range <>) of cable_rec;

  type cable_ptr is access cable_vec;

  type membr_rec is
    record
      stiffness : long_float; -- Young's modulus * thickness
      poisson   : long_float; -- Poisson's ratio
      area      : long_float;
      l1        : long_float;
      l2        : long_float;
      l3        : long_float;
      glob      : terms.imat1 (1 .. 3);
      u1        : terms.fmat1 (1 .. 3);
      u2        : terms.fmat1 (1 .. 3);
      u3        : terms.fmat1 (1 .. 3);
      h         : terms.fmat2 (1 .. 3, 1 .. 3);
    end record;

  type membr_vec is array (integer range <>) of membr_rec;

  type membr_ptr is access membr_vec;

  optimal,loading      : boolean;
  nn,n,maxlist         : integer;
  maxls,minor,maxminor : integer;
  lstol,mntol,mnref    : long_float;
  maxnorm,f            : long_float;
  index                : terms.imat2_ptr;
  xo,yo,zo,x,g         : terms.fmat1_ptr;
  known,force,action  : terms.fmat2_ptr;
  cable                : cable_ptr;
  membr                : membr_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  procedure putchar (io : in text_io.file_type;
                     c  : in character;
                     n  : in integer) is

  begin
    for k in 1 .. n loop
      text_io.put(io,c);
    end loop;
  end putchar;

  procedure cableobj (cable                   : in cable_rec;
                      epsilon,strain,axial,fe : out long_float;
                      x                       : in terms.fmat1_ptr;
                      ge                      : out fmat2x3) is

    use
      ada.numerics.long_elementary_functions;

    glob,i          : integer;
    c,l,alpha,normw : long_float;
    z,w             : terms.fmat1 (1 .. 3);
    xe              : terms.fmat2 (1 .. 2, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 2 loop
      glob := cable.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- calculate z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / cable.length;
    end loop;

    -- calculate w
    normw := 0.0;
    for axis in 1 .. 3 loop
      w(axis) := cable.u(axis) + z(axis);
      normw := normw + w(axis) ** 2;
    end loop;
    normw := sqrt(normw);

    -- calculate alpha
    alpha := 0.0;
    for axis in 1 .. 3 loop
      alpha := alpha + z(axis) * (cable.u(axis) + w(axis));
    end loop;

    -- calculate epsilon, strain, axial, fe
    epsilon := alpha / (normw + 1.0);

    case cable.state is
      -- normal state
      when 0 =>
        strain := epsilon;
        axial := cable.young * strain * cable.area;
        fe := 0.5 * axial * strain * cable.length;

      -- cut state
      when 1 =>
        c := cable.cut / cable.length;
        l := (1.0 - c);
        strain := (epsilon + c) / l;
        axial := cable.young * strain * cable.area;
        fe := 0.5 * axial * strain * cable.length * l;

      -- tension state
      when 2 =>
        strain := cable.tension / cable.young;
        axial := cable.tension * cable.area;
        fe := cable.length * axial * normw;

      when others =>
        null;
    end case;

    -- global gradient
    c := axial / normw;
    for axis in 1 .. 3 loop
      alpha := c * w(axis);
      ge(1,axis) := -alpha;
      ge(2,axis) :=  alpha;
    end loop;
  end cableobj;

  procedure membrobj (membr          : in membr_rec;
                      st1,st2,st3,fe : out long_float;
                      x              : in terms.fmat1_ptr;
                      gme            : out fmat9x1) is

    glob,i                      : integer;
    delta1,e1,e2,e3,sv1,sv2,sv3 : long_float;
    z                           : terms.fmat1 (1 .. 3);
    de1,de2,de3                 : terms.fmat1 (1 .. 9);
    xe                          : terms.fmat2 (1 .. 3, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 3 loop
      glob := membr.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- side 1

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(3,axis) - xe(2,axis)) / membr.l1;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (membr.u1(axis) + z(axis)) ** 2;
    end loop;

    -- e1
    e1 := (delta1 - 1.0) / 2.0;

    -- de1
    de1(1) := 0.0;
    de1(2) := 0.0;
    de1(3) := 0.0;

    de1(7) := (membr.u1(1) + z(1)) / membr.l1;
    de1(8) := (membr.u1(2) + z(2)) / membr.l1;
    de1(9) := (membr.u1(3) + z(3)) / membr.l1;

    de1(4) := -de1(7);
    de1(5) := -de1(8);
    de1(6) := -de1(9);

    -- side 2

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(1,axis) - xe(3,axis)) / membr.l2;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (membr.u2(axis) + z(axis)) ** 2;
    end loop;

    -- e2
    e2 := (delta1 - 1.0) / 2.0;

    -- de2
    de2(4) := 0.0;
    de2(5) := 0.0;
    de2(6) := 0.0;

    de2(1) := (membr.u2(1) + z(1)) / membr.l2;
    de2(2) := (membr.u2(2) + z(2)) / membr.l2;
    de2(3) := (membr.u2(3) + z(3)) / membr.l2;

    de2(7) := -de2(1);
    de2(8) := -de2(2);
    de2(9) := -de2(3);

    -- side 3

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / membr.l3;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (membr.u3(axis) + z(axis)) ** 2;
    end loop;

    -- e3
    e3 := (delta1 - 1.0) / 2.0;

    -- de3
    de3(7) := 0.0;
    de3(8) := 0.0;
    de3(9) := 0.0;

    de3(4) := (membr.u3(1) + z(1)) / membr.l3;
    de3(5) := (membr.u3(2) + z(2)) / membr.l3;
    de3(6) := (membr.u3(3) + z(3)) / membr.l3;

    de3(1) := -de3(4);
    de3(2) := -de3(5);
    de3(3) := -de3(6);

    -- stress * thickness
    st1 := membr.h(1,1) * e1 + membr.h(1,2) * e2 + membr.h(1,3) * e3;
    st2 := membr.h(2,1) * e1 + membr.h(2,2) * e2 + membr.h(2,3) * e3;
    st3 := membr.h(3,1) * e1 + membr.h(3,2) * e2 + membr.h(3,3) * e3;

    -- stress * volume
    sv1 := st1 * membr.area;
    sv2 := st2 * membr.area;
    sv3 := st3 * membr.area;

    -- residue
    for i in 1 .. 9 loop
      gme(i) := sv1 * de1(i) + sv2 * de2(i) + sv3 * de3(i);
    end loop;

    -- energy
    fe := (sv1 * e1 + sv2 * e2 + sv3 * e3) / 2.0;
  end membrobj;

  procedure setstate (e     : in integer;
                      state : in integer) is

    epsilon,strain,axial,fe : long_float;
    gce                     : fmat2x3;

  begin
    case state is
      -- set normal state
      when 0 =>
        cable(e).cut := 0.0;
        cable(e).tension := 0.0;
        cable(e).state := 0;

      -- set cut state with current tension value
      when 1 =>
        cableobj(cable(e),epsilon,strain,axial,fe,x,gce);
        cable(e).cut := (strain - epsilon) * cable(e).length / (1.0 + strain);
        cable(e).state := 1;

      -- set tension state with current cut value
      when 2 =>
        cableobj(cable(e),epsilon,strain,axial,fe,x,gce);
        cable(e).tension := strain * cable(e).young;
        cable(e).state := 2;

      when others =>
        null;
    end case;
  end setstate;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

    glob,i,s             : integer;
    epsilon,strain,axial : long_float;
    action,fe            : long_float;
    st1,st2,st3          : long_float;
    gce                  : fmat2x3;
    gme                  : fmat9x1;

  begin
    -- initiate objective, gradient
    f := 0.0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i <= n then
          action := force(glob,axis);
          f := f - action * x(i);
          g(i) := -action;
        end if;
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gce);

      -- update objective
      f := f + fe;

      -- update gradient
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gce(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;

    -- membrane elements
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gme);

      -- update objective
      f := f + fe;

      -- update gradient
      s := 0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gme(s);
          end if;
        end loop;
      end loop;
    end loop;
  end objective;

  package qn is new nlp.qnewton(objective);

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    e,glob,axis,state : integer;
    ne,nk             : integer;
    io                : file_type;

  begin
    open(io, in_file, "input.txt");

    -- solution parameters
    scan(io);
    get(io,lstol);
    get(io,maxls);
    get(io,mntol);
    get(io,mnref);
    get(io,maxminor);
    get(io,maxlist);
    skip_line(io);

    if mntol <= 0.0 then
      mntol := 0.001;
    end if;

    -- node coordinates
    scan(io);
    get(io,nn);
    skip_line(io);

    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    skip_line(io);
    for k in 1 .. nn loop
      get(io,glob);
      get(io,xo(glob));
      get(io,yo(glob));
      get(io,zo(glob));
      skip_line(io);
    end loop;

    -- cable elements
    scan(io);
    get(io,ne);
    skip_line(io);

    cable := new cable_vec (1 .. ne);

    if ne > 0 then
      skip_line(io);
      for k in cable'range loop
        get(io,e);
        get(io,cable(e).glob(1));
        get(io,cable(e).glob(2));
        get(io,cable(e).area);
        get(io,cable(e).young);
        skip_line(io);
      end loop;
    end if;

    -- membrane elements
    scan(io);
    get(io,ne);
    skip_line(io);

    membr := new membr_vec (1 .. ne);

    if ne > 0 then
      skip_line(io);
      for k in membr'range loop
        get(io,e);
        get(io,membr(e).glob(1));
        get(io,membr(e).glob(2));
        get(io,membr(e).glob(3));
        get(io,membr(e).stiffness);
        get(io,membr(e).poisson);
        skip_line(io);
      end loop;
    end if;

    -- initiate index
    index := new terms.imat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        index(glob,axis) := 0;
      end loop;
    end loop;

    -- known displacements
    known := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        known(glob,axis) := 0.0;
      end loop;
    end loop;

    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,known(glob,axis));
        skip_line(io);
        index(glob,axis) := 1;
      end loop;
    end if;

    -- set cable element state
    scan(io);
    get(io,ne);
    skip_line(io);

    if ne > 0 then
      skip_line(io);

      for k in 1 .. ne loop
        get(io,e);
        get(io,state);

        case state is
          -- set cut state with given cut value
          when 1 =>
            get(io,cable(e).cut);
            cable(e).state := 1;

          -- set tension state with given tension value
          when 2 =>
            get(io,cable(e).tension);
            cable(e).state := 2;

          when others =>
            null;
        end case;

        skip_line(io);
      end loop;
    end if;

    -- action
    scan(io);
    get(io,ne);
    skip_line(io);

    loading := ne > 0;

    if loading then
      action := new terms.fmat2 (1 .. nn, 1 .. 3);
      for glob in 1 .. nn loop
        for axis in 1 .. 3 loop
          action(glob,axis) := 0.0;
        end loop;
      end loop;

      skip_line(io);
      for e in 1 .. ne loop
        get(io,glob);
        get(io,axis);
        get(io,action(glob,axis));
        skip_line(io);
      end loop;
    end if;

    close(io);
  end input;

  procedure setup is

    use
      ada.numerics.long_elementary_functions;

    glob1,glob2,glob3,s1,s2       : integer;
    aux,norm,poisson              : long_float;
    cos1,cos2,cos3,sin1,sin2,sin3 : long_float;
    u                             : terms.fmat1 (1 .. 3);

  begin
    -- number of unknowns
    n := 0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          n := n + 1;
        end if;
      end loop;
    end loop;

    -- index
    s1 := 0;
    s2 := n;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          s1 := s1 + 1;
          index(glob,axis) := s1;
        else
          s2 := s2 + 1;
          index(glob,axis) := s2;
        end if;
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      glob1 := cable(e).glob(1);
      glob2 := cable(e).glob(2);

      u(1) := xo(glob2) - xo(glob1);
      u(2) := yo(glob2) - yo(glob1);
      u(3) := zo(glob2) - zo(glob1);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);

      cable(e).length := norm;
      cable(e).u(1) := u(1) / norm;
      cable(e).u(2) := u(2) / norm;
      cable(e).u(3) := u(3) / norm;
    end loop;

    -- membrane elements
    for e in membr'range loop
      glob1 := membr(e).glob(1);
      glob2 := membr(e).glob(2);
      glob3 := membr(e).glob(3);

      -- u1
      u(1) := xo(glob3) - xo(glob2);
      u(2) := yo(glob3) - yo(glob2);
      u(3) := zo(glob3) - zo(glob2);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        membr(e).u1(axis) := u(axis) / norm;
      end loop;
      membr(e).l1 := norm;

      -- u2
      u(1) := xo(glob1) - xo(glob3);
      u(2) := yo(glob1) - yo(glob3);
      u(3) := zo(glob1) - zo(glob3);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        membr(e).u2(axis) := u(axis) / norm;
      end loop;
      membr(e).l2 := norm;

      -- u3
      u(1) := xo(glob2) - xo(glob1);
      u(2) := yo(glob2) - yo(glob1);
      u(3) := zo(glob2) - zo(glob1);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        membr(e).u3(axis) := u(axis) / norm;
      end loop;
      membr(e).l3 := norm;

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - membr(e).u2(axis) * membr(e).u3(axis);
      end loop;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - membr(e).u3(axis) * membr(e).u1(axis);
      end loop;

      -- cosine alpha3
      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 - membr(e).u1(axis) * membr(e).u2(axis);
      end loop;

      -- sine alpha1
      sin1 := sqrt(1.0 - cos1 ** 2);

      -- sine alpha2
      sin2 := sqrt(1.0 - cos2 ** 2);

      -- sine alpha3
      sin3 := sqrt(1.0 - cos3 ** 2);

      -- h
      poisson := membr(e).poisson;
      aux := membr(e).stiffness / (1.0 - poisson ** 2) / 2.0;
      membr(e).h(1,1) := aux * (cos1 ** 2 + 1.0 - poisson * sin1 ** 2) / sin2 ** 2 / sin3 ** 2;
      membr(e).h(1,2) := aux * (cos1 * cos2 - cos3 + poisson * sin1 * sin2) / sin1 / sin2 / sin3 ** 2;
      membr(e).h(1,3) := aux * (cos1 * cos3 - cos2 + poisson * sin1 * sin3) / sin1 / sin2 ** 2 / sin3;
      membr(e).h(2,2) := aux * (cos2 ** 2 + 1.0 - poisson * sin2 ** 2) / sin1 ** 2 / sin3 ** 2;
      membr(e).h(2,3) := aux * (cos2 * cos3 - cos1 + poisson * sin2 * sin3) / sin1 ** 2 / sin2 / sin3;
      membr(e).h(3,3) := aux * (cos3 ** 2 + 1.0 - poisson * sin3 ** 2) / sin1 ** 2 / sin2 ** 2;
      membr(e).h(2,1) := membr(e).h(1,2);
      membr(e).h(3,1) := membr(e).h(1,3);
      membr(e).h(3,2) := membr(e).h(2,3);

      -- area
      membr(e).area := membr(e).l1 * membr(e).l2 * sin3 / 2.0;
    end loop;

    -- initiate x
    x := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      x(k) := 0.0;
    end loop;

    -- initiate g
    g := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      g(k) := 0.0;
    end loop;
  end setup;

  procedure vertex (glob : in integer;
                    vx   : out fmat3x1) is

    i : integer;

  begin
    vx(1) := xo(glob);
    vx(2) := yo(glob);
    vx(3) := zo(glob);

    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        vx(axis) := vx(axis) + x(i);
      else
        vx(axis) := vx(axis) + known(glob,axis);
      end if;
    end loop;
  end vertex;

  procedure output0 is

    use
      text_io;

    io : file_type;

  begin
    create(io, out_file, "output.txt");

    put(io,"Tensegrity");
    new_line(io);

    close(io);
  end output0;

  procedure output (loadname : in integer) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    glob,s,last              : integer;
    epsilon,strain,axial,fe  : long_float;
    st1,st2,st3              : long_float;
    s1,s2,cos1,cos2,cos3,aux : long_float;
    gce                      : fmat2x3;
    gme                      : fmat9x1;
    v                        : fmat3x1;
    io                       : file_type;

  begin
    open(io, append_file, "output.txt");

    new_line(io);
    new_line(io);

    putchar(io,'-',41);
    new_line(io);

    case loadname is
      when 0 =>
        put(io,"PRESTRESS");
        new_line(io);

      when 1 =>
        put(io,"LOADING");
        new_line(io);

      when others =>
        null;
    end case;

    put(io,"Minor Iterations          = ");
    put(io,minor,13);
    new_line(io);
    put(io,"Total Potential Energy    = ");
    put(io,f,2,6,3);
    new_line(io);
    put(io,"Allowable Maximum Residue = ");
    put(io,maxnorm,2,6,3);
    new_line(io);

    putchar(io,'-',41);
    new_line(io);

    -- coordinates
    new_line(io);
    new_line(io);

    putchar(io,'-',59);
    new_line(io);
    put(io," Node           Coord X           Coord Y           Coord Z");
    new_line(io);
    putchar(io,'-',59);
    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,v);
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        put(io,v(axis),2,6,3);
      end loop;
      new_line(io);
    end loop;

    -- cable elements
    new_line(io);
    new_line(io);

    putchar(io,'-',23);
    new_line(io);
    put(io," Elem             Force");
    new_line(io);
    putchar(io,'-',23);
    new_line(io);

    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gce);
      put(io,e,5);
      putchar(io,' ',5);
      put(io,axial,2,6,3);
      new_line(io);
    end loop;

    -- membrane elements
    new_line(io);
    new_line(io);

    putchar(io,'-',41);
    new_line(io);
    put(io," Elem          Stress-1          Stress-2");
    new_line(io);
    putchar(io,'-',41);
    new_line(io);

    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gme);

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - membr(e).u2(axis) * membr(e).u3(axis);
      end loop;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - membr(e).u3(axis) * membr(e).u1(axis);
      end loop;

      -- cosine alpha3
      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 - membr(e).u1(axis) * membr(e).u2(axis);
      end loop;

      -- principal stresses
      aux := 0.0;
      aux := aux + st1 * st2 * (cos3 ** 2 - 0.5);
      aux := aux + st1 * st3 * (cos2 ** 2 - 0.5);
      aux := aux + st2 * st3 * (cos1 ** 2 - 0.5);
      aux := 4.0 * aux + st1 ** 2 + st2 ** 2 + st3 ** 2;
      aux := sqrt(aux);
      s1 := ((st1 + st2 + st3) + aux) / 2.0;
      s2 := ((st1 + st2 + st3) - aux) / 2.0;

      put(io,e,5);
      putchar(io,' ',5);
      put(io,s1,2,6,3);
      putchar(io,' ',5);
      put(io,s2,2,6,3);
      new_line(io);
    end loop;

    -- initiate support reactions
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := -force(glob,axis);
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gce);

      -- update support reactions
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          if index(glob,axis) > n then
            force(glob,axis) := force(glob,axis) + gce(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;

    -- membrane elements
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gme);

      -- update support reactions
      s := 0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          force(glob,axis) := force(glob,axis) + gme(s);
        end loop;
      end loop;
    end loop;

    -- support reactions
    new_line(io);
    new_line(io);

    putchar(io,'-',59);
    new_line(io);
    put(io," Node           Force X           Force Y           Force Z");
    new_line(io);
    putchar(io,'-',59);
    new_line(io);

    for glob in 1 .. nn loop
      last := 0;
      for axis in 1 .. 3 loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last := axis;
        end if;
      end loop;

      if last > 0 then
        put(io,glob,5);
        for axis in 1 .. last loop
          if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
            putchar(io,' ',5);
            put(io,force(glob,axis),2,6,3);
          else
            putchar(io,' ',18);
          end if;
        end loop;
        new_line(io);
      end if;
    end loop;

    close(io);
  end output;

  procedure autocad0 is

    use
      text_io,
      ada.long_float_text_io;

    absent : boolean;
    e,glob : integer;
    io     : file_type;

  begin
    create(io, out_file, "autocad.scr");

    -- autocad
    put(io,"osmode 0");
    new_line(io);

    put(io,"layer make Geometry ");
    new_line(io);

    -- cable elements
    e := 0;
    absent := true;
    while e < cable'last and absent loop
      e := e + 1;
      absent := cable(e).state = 0;
    end loop;

    if not absent then
      put(io,"-color");
      new_line(io);
      put(io,"green");
      new_line(io);

      for e in cable'range loop
        if cable(e).state /= 0 then
          put(io,"3dpoly ");
          for loc in 1 .. 2 loop
            glob := cable(e).glob(loc);
            put(io,xo(glob),1,8,3);
            put(io,",");
            put(io,yo(glob),1,8,3);
            put(io,",");
            put(io,zo(glob),1,8,3);
            put(io," ");
          end loop;
          new_line(io);
        end if;
      end loop;
    end if;

    -- cable elements
    e := 0;
    absent := true;
    while e < cable'last and absent loop
      e := e + 1;
      absent := cable(e).state /= 0;
    end loop;

    if not absent then
      put(io,"-color");
      new_line(io);
      put(io,"white");
      new_line(io);

      for e in cable'range loop
        if cable(e).state = 0 then
          put(io,"3dpoly ");
          for loc in 1 .. 2 loop
            glob := cable(e).glob(loc);
            put(io,xo(glob),1,8,3);
            put(io,",");
            put(io,yo(glob),1,8,3);
            put(io,",");
            put(io,zo(glob),1,8,3);
            put(io," ");
          end loop;
          new_line(io);
        end if;
      end loop;
    end if;

    -- membrane elements
    if membr'last > 0 then
      put(io,"-color");
      new_line(io);
      put(io,"white");
      new_line(io);

      for e in membr'range loop
        put(io,"3dface ");
        for loc in 1 .. 3 loop
          glob := membr(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        put(io," ");
        new_line(io);
      end loop;
    end if;

    -- autocad
    put(io,"-color");
    new_line(io);
    put(io,"white");
    new_line(io);

    put(io,"layer set 0 ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad0;

  procedure autocad (loadname : in integer) is

    use
      text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    glob                     : integer;
    epsilon,strain,axial,fe  : long_float;
    st1,st2,st3              : long_float;
    s1,s2,cos1,cos2,cos3,aux : long_float;
    gce                      : fmat2x3;
    gme                      : fmat9x1;
    vx                       : fmat3x1;
    io                       : file_type;

  begin
    open(io, append_file, "autocad.scr");

    case loadname is
      when 0 =>
        put(io,"layer make Prestress ");
        new_line(io);

      when 1 =>
        put(io,"layer make Loading ");
        new_line(io);

      when others =>
        null;
    end case;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gce);

      if axial < 0.0 then
        put(io,"-color");
        new_line(io);
        put(io,"red");
        new_line(io);
      else
        put(io,"-color");
        new_line(io);
        put(io,"blue");
        new_line(io);
      end if;

      put(io,"3dpoly ");
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        vertex(glob,vx);
        put(io,vx(1),1,8,3);
        put(io,",");
        put(io,vx(2),1,8,3);
        put(io,",");
        put(io,vx(3),1,8,3);
        put(io," ");
      end loop;
      new_line(io);
    end loop;

    -- membrane elements
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gme);

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - membr(e).u2(axis) * membr(e).u3(axis);
      end loop;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - membr(e).u3(axis) * membr(e).u1(axis);
      end loop;

      -- cosine alpha3
      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 - membr(e).u1(axis) * membr(e).u2(axis);
      end loop;

      -- principal stresses
      aux := 0.0;
      aux := aux + st1 * st2 * (cos3 ** 2 - 0.5);
      aux := aux + st1 * st3 * (cos2 ** 2 - 0.5);
      aux := aux + st2 * st3 * (cos1 ** 2 - 0.5);
      aux := 4.0 * aux + st1 ** 2 + st2 ** 2 + st3 ** 2;
      aux := sqrt(aux);
      s1 := ((st1 + st2 + st3) + aux) / 2.0;
      s2 := ((st1 + st2 + st3) - aux) / 2.0;

      if s1 < 0.0 or s2 < 0.0 then
        put(io,"-color");
        new_line(io);
        put(io,"red");
        new_line(io);
      else
        put(io,"-color");
        new_line(io);
        put(io,"blue");
        new_line(io);
      end if;

      put(io,"3dface ");
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        vertex(glob,vx);
        put(io,vx(1),1,8,3);
        put(io,",");
        put(io,vx(2),1,8,3);
        put(io,",");
        put(io,vx(3),1,8,3);
        put(io," ");
      end loop;
      put(io," ");
      new_line(io);
    end loop;

    -- autocad
    put(io,"-color");
    new_line(io);
    put(io,"white");
    new_line(io);

    put(io,"layer set 0 ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    case loadname is
      when 0 =>
        put(io,"layer off Prestress ");
        new_line(io);

      when 1 =>
        put(io,"layer off Loading ");
        new_line(io);

      when others =>
        null;
    end case;

    close(io);
  end autocad;

begin
  input;

  setup;

  output0;

  autocad0;

  -- nodal forces
  force := new terms.fmat2 (1 .. nn, 1 .. 3);
  for glob in 1 .. nn loop
    for axis in 1 .. 3 loop
      force(glob,axis) := 0.0;
    end loop;
  end loop;

  objective(n,f,x,g);

  -- maxnorm
  if mnref <= 0.0 then
    maxnorm := mntol * nlp.norm(n,g);
  else
    maxnorm := mntol * mnref;
  end if;

  qn.lbfgs(maxlist,n,maxminor,maxls,
           minor,
           maxnorm,lstol,
           f,
           x,g,
           optimal);

  output(0);

  autocad(0);

  if loading then
    -- set cut state with current tension value
    for e in cable'range loop
      if cable(e).state = 2 then
        setstate(e,1);
      end if;
    end loop;

    -- nodal forces
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := action(glob,axis);
      end loop;
    end loop;

    objective(n,f,x,g);

    -- maxnorm
    if mnref <= 0.0 then
      maxnorm := mntol * nlp.norm(n,g);
    else
      maxnorm := mntol * mnref;
    end if;

    qn.lbfgs(maxlist,n,maxminor,maxls,
             minor,
             maxnorm,lstol,
             f,
             x,g,
             optimal);

    output(1);

    autocad(1);
  end if;
end tensegrity;
